% CheckCalcs.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011-2013  Wolfram Kahl
% All rights reserved
%
\section{|CheckCalcs|: Checking Calculations}

\begin{code}
{-# LANGUAGE PatternGuards #-}
module CheckCalcs where

import qualified AC
import AC  ( Term(Appl, AC), unVar, TermPair, applyACs, applyCs, apply
           -- |, applySubst|
           , removeSelfInverse )
import Rewrite
import LogicE
import Utils.Q
import Utils.List (groupFst)

import TeXParse
import qualified Parser as P (expr)
import Text.ParserCombinators.Parsec hiding (Parser)

import qualified Data.Map as Map
import qualified Data.Char as Char
import Data.List (partition, nub, sort)
import Data.Maybe (maybeToList, catMaybes)

import Control.Monad (mplus)
import Control.Arrow (first, second)
\end{code}


%{{{ doSubst
|doSubst| is an auxiliary function for |trySubst|.
\begin{code}
doSubst :: Term Var Op Const -> [Term Var Op Const]
doSubst (Appl '⇇' ts) = case splitSubstArgs ts of
  Left msg -> error $ "Internal error, please report: doSubst: " ++ msg
  Right (e,(es,vs0)) -> case mapM unVar vs0 of
    Nothing -> error $ "Internal error, please report: doSubst: non-variables! " ++ show vs0 ++ " := " ++ show es
    Just vs -> let su = Map.fromList $ zip vs es
               in maybeToList $ applySubstE su e
doSubst _ = []
\end{code}
%}}}

%{{{ applySubstE
Since substitution as implemented now
does not create fresh variables,
it is a partial operation.
\begin{code}
applySubstE :: AC.Subst Var Op Const -> Term Var Op Const -> Maybe (Term Var Op Const)
applySubstE su = qDoMaybe q
  where
    q :: Term Var Op Const -> Maybe (Maybe (Term Var Op Const))
    q t@(AC.Var v) = case Map.lookup v su of
      Just t' | t == t' -> Just Nothing
      m -> Just m
    q (AC.Const _) = Just Nothing
    q (Appl '⇇' ts) =  case splitSubstArgs ts of
       Left msg -> error $ "Internal error, please report: applySubstE: " ++ msg
       Right (_e,(_es,_vs0)) -> Nothing -- refusing to substitute into substitution applications, for the time being
               -- |fmap  (\ es' -> AC.apply '⇇' (vs0 ++ es' ++ [e]))|
               -- |      (qList q es)|
    -- \edcomm{WK}{Don't substitute into quantifications yet:}
    q (AC.Appl op [_vs, _mr, _e]) | Just _bOp <- quantToBin op = Nothing
    q (AC.Appl op ts) = fmap (fmap (apply op)) $ qListMaybe q ts
    q (AC.C op t1 t2) = fmap (fmap (applyCs op)) $ qListMaybe q [t1, t2]
    q (AC.AC op ts) = fmap (fmap (AC.applyACs op)) $ qListMaybe q ts
\end{code}
%}}}

\edcomm{WK}{Why no |q|?}
\begin{code}
trySubst :: Term Var Op Const -> [Term Var Op Const]
trySubst t@(AC.Var _) = [t]
trySubst t@(AC.Const _) = [t]
trySubst t@(Appl op ts) = map (Appl op) (mapM trySubst ts) ++ (doSubst t >>= trySubst)
trySubst (AC.C op t1 t2) = map (applyCs op) (mapM trySubst [t1,t2])
trySubst (AC op ts) = map (applyACs op) (mapM trySubst ts)
\end{code}


\begin{code}
checkItems :: [Item] -> [Item]
checkItems = map . mapCalcs $ map checkCalc
\end{code}

\savecolumns
\begin{code}
checkCalc :: Calc -> Calc
checkCalc ca = ca {calcGoalCheck = nokStepsR ++ gCheck, calcSteps = steps'}
  where
    g = calcGoal ca
    e = calcStart ca
    steps = calcSteps ca
    steps' = h (toTerm e) steps
    stepRs :: [Report]
    stepRs = map calcCheck steps'
    nokStepsR :: [Report]
    nokStepsR = case filter (not . confirmedOK) stepRs of
      [] -> if null stepRs then [] else [Report ["All steps OK"] True]
      noks -> [Report [shows (length noks) " out of " ++ shows (length stepRs) " steps not justified"] False]
    getGoal :: CalcGoal -> [Expr]
    getGoal gg = case goalExpr gg of
          Just x -> [x]
          Nothing -> case goalNr gg of
            Just nr -> map snd $ findThNum nr
            Nothing -> []
    gCheck :: [Report]
    gCheck = let
        e' = calcResult $ last steps
        matchThisIs th expr = do
           th' <- equEquiv $ toTerm th
           AC.match freshVars [(th', toTerm expr)]
        (checkThisIs, other) = let
             finalThisIs = "Final \\ThisIs{" ++ showsGoal (calcEndDecl ca) "}"
             initialThisIs = "Initial \\ThisIs{" ++ showsGoal (calcStartDecl ca) "}"
           in if isNoGoal $ calcStartDecl ca
             then if isNoGoal $ calcEndDecl ca
               then (Nothing, [])
               else case getGoal $ calcEndDecl ca of
                 [] -> (Just (Report ["Final \\ThisIs{...} not understood"] False),[])
                 ees -> case ees >>= flip matchThisIs e' of
                   [] -> (Just (Report [finalThisIs ++ " does not match"] False),[])
                   _ -> (Just (Report [finalThisIs ++ " matches ─ OK"] True), [e])
             else if isNoGoal $ calcEndDecl ca
               then case getGoal $ calcStartDecl ca of
                 [] -> (Just (Report ["Initial \\ThisIs{...} not understood"] False),[])
                 ees -> case ees >>= flip matchThisIs e of
                   [] -> (Just (Report [initialThisIs ++ " does not match"] False),[])
                   _ -> (Just (Report [initialThisIs ++ " matches ─ OK"] True), [e'])
               else (Just (Report ["Only one \\ThisIs{...} allowed!"] False),[])
\end{code}

\restorecolumns
\begin{code}
      in case conjOps $ map calcOp steps of
        Nothing -> []
        Just op -> let
          rs = map toTerm $ Bin op e e' : case op of
            '≡' -> other ++ Bin op e' e : case filter (not . isTrue) [e, e'] of
                     [e''] -> [e'']
                     _ -> []
            '=' -> other ++ Bin '=' e' e : Bin '≡' e e' : Bin '≡' e' e : case filter (not . isTrue) [e, e'] of
                     [e''] -> [e'']
                     _ -> []
            '⇒' -> Bin '⇐' e' e : if isTrue e then [e'] else []
            '⇐' -> Bin '⇒' e' e : if isTrue e' then [e] else []
            _ -> []
\end{code}

\restorecolumns
\begin{code}
          gc :: Maybe Report
          gc = case nub (map toTerm $ getGoal g) of
            [] -> if isNoGoal g then Nothing
                  else Just (Report ["Could not extract information from stated goal."] False)
            [g'] -> if g' `elem` rs -- \edcomm{WK}{convert |ConjChain|s!}
              then Just (Report ["Proof matches goal ─ OK"] True)
              else Just (Report ["Could not match proof with stated goal."] False)
            _ -> Just (Report ["Ambigous theorem reference in stated goal."] False)
          in catMaybes [checkThisIs, gc] -- s ++ "\n─  " ++ s'
        -- \edcomm{WK}{Additional sanity checking required!}
\end{code}

\restorecolumns
\begin{code}
    h _ [] = []
    h t (c : cs) = let
      t' = toTerm (calcResult c)
      ((comment, (saysSubst, saysParen)), rs) = hintRules c
      (rs1, rs2) = partition isSymEquiv rs
      rs1' = case rs1 of
        [] -> id
        (r : _) -> (:) (map ((,) ((Entry symEquivTP [] r emptyNOR) :)) . removeSelfInverse ('≡' ==))
      rs' :: [Rewrite Var Op Const Theorem]
      rs' = rs1' . map (rewrite'' unVarList freshVars) $ mkRules rs2 -- \edcomm{WK}{Add double negation etc.?}
      s = case     check trySubst [rs',rs'] (t, t')
          `mplus`  check trySubst [rs',rs'] (t', t) of
          Nothing -> if t == t'
            then if null rs && saysParen
                 then Report ["Parentheses  ─   OK"] True
                 else Report [comment ++ "   ─   OK (no change)"] True
            else if any (`elem` (trySubst t')) (trySubst t)
                 then if null rs && saysSubst
                      then Report ["Substitution   ─   OK"] True
                      else Report [comment ++ "   ─   OK (substitution)"] True
                 else Report [comment ++ "; could not justify this step!"] False
          Just qs -> prependReport comment $ let
               nors = map e_nor qs
               (ierrs,ps) = unzip nors
               (occs, oks) = cleanOccsNOccs ps
             in case concat ierrs of
               [] -> case occs of
                  [] -> Report [ (if null oks then id else breakA) "   ─   OK"
                                  ++ case oks of
                                     [] -> ""
                                     oks' -> ": " ++ showsNotOccurs' oks' ""
                               ] True
                  occs' -> Report ["   ─   proviso violated: " ++
                              showsCommaSep1 (map (\ (v,ee) -> showsOccurs (shows (Var v)) (shows ee)) $ occs') ""] False
               ierrs' -> Report ("   ─   INTERNAL ERROR:" : nub ierrs') False
            -- |concatMap (\ ((l,r),rule) -> head (fst $ fst rule) ++ ' ' : show (apply '=' [l,r]) ++ ". ") qs|
      in c {calcCheck = s} : h t' cs
\end{code}

\begin{code}
showsNotOccurs' :: (Show t, Ord t) => [(Var, t)] -> ShowS
showsNotOccurs' ps = let
    sh (es, vs) = showsNotOccurs  (showsCommaSep1 (map (shows . Var) vs))
                                  (showsCommaSep1 (map shows es))
    swap (x,y) = (y,x)
  in showsCommaSep1 (map sh . groupFst . sort . map swap . groupFst $ sort ps)
\end{code}

\begin{code}
showRule :: (([String],[String]), Expr) -> String
showRule ((ns,ds),e) = head ns ++ ' ' : case ds of
  [] -> show e
  d : _ -> d

showRules :: String -> [(([String],[String]), Expr)] -> String
showRules sep rs = showsSep (sep ++) (map (++) . nub $ map showRule rs) ""
\end{code}
\begin{verbatim}
{-
showRules _sep [] = ""
showRules sep (r : rs) = h (showRule r) rs
  where
    h s [] = s
    h s (r' : rs') = s ++ sep ++ h (showRule r') rs'
-}
\end{verbatim}
 
\begin{code}
hintRules :: CalcStep -> ((String , (Bool, Bool)), [Theorem])
hintRules step = case calcOp step of
  '=' -> case parseHint (calcOp step) $ show $ calcHint step of
    Left _ -> (ff "Could not extract information", [])
    Right [] -> (ff "Could not extract information", [])
    Right his -> case sepHI his of
      (ss , []) -> (("Could not extract theorem information", ss), [])
      (ss , rs) -> ((showRules ", " rs, ss), rs)
  op -> (ff ("Cannot handle CalcStep operator ‶" ++ op : "″ yet"), [])
  where ff s = (s, (False, False))
\end{code}

\begin{code}
data HintInfo
  = THM Theorem
  | SUBST
  | PAREN
\end{code}

\begin{code}
sepHI :: [HintInfo] -> ((Bool , Bool) , [Theorem])
sepHI [] = ((False , False) , [])
sepHI (SUBST : his) = first (first (const True)) (sepHI his)
sepHI (PAREN : his) = first (second (const True)) (sepHI his)
sepHI (THM t : his) = second ((:) t) (sepHI his)
\end{code}

\begin{code}
parseHint :: Char -> String -> Either ParseError [HintInfo]
parseHint '=' s = parse pHint "" s
parseHint _ _s = Right []
\end{code}

\begin{code}
pHint, pHint1 :: Parser st [HintInfo]
pHint  = concat <$> many pHint1
pHint1  = fmap (map THM) pThNum
       <|> (fmap snd (pTeXMathWrap pMath) >>= hintMath)
       <|> brace pHint  -- |brace'| is inappropriate here.
       <|> onlyWrap pHint
       <|> otherHint
  where
    pMath = many1 $ satisfy (\ c -> c /= '$') -- \unfinished
    hintMath s = case parse (P.expr >>= return . findExprInLaws) "" s of
      Left _ -> return [] --- |fail| wold block later alternatives.
      Right ts -> return $ map (THM . fst) ts

-- \edcomm{WK}{Should the substitutions be applied so that only the named instance can be used?}
findExprInLaws :: Expr -> [(Theorem, Term Var Op Const)]
findExprInLaws e = case toTerm e of
   e' -> filter (\ (_ , t) -> not . null $ AC.match freshVars [(t,e')]) laws'

otherHint :: Parser st [HintInfo]
otherHint
  =    try (oneOf "sS" >> string "ubst" >> trailing >> return [SUBST])
  <|>  try (oneOf "pP" >> string "aren" >> trailing >> return [PAREN])
  <|>  (hintSkip >> return [])
  where
    trailing = many letter >> notFollowedBy digit

hintSkip  :: Parser st String
hintSkip  = many1 alphaNum
          <|> texspace
          <|> pMacro0  -- \edcomm{WK}{Why is this allowed?}
          <|> fmap (: []) (satisfy (\ c -> Char.isSeparator c || (notElem c "\\{}$" && Char.isPunctuation c)))
\end{code}

\begin{code}
pThNum :: Parser st [Theorem]
pThNum = findThNum <$> try parenThNum
\end{code}

\begin{code}
findThNum :: String -> [Theorem]
findThNum s =  let n = '(' : s ++ ")"
               in filter (\ ((ns, _ds), _e) -> n `elem` ns) laws
\end{code}

\begin{code}
\end{code}

\begin{code}
thms :: String -> [Theorem]
thms s = case parse pThNum "" s of
  Left _ -> []
  Right ts -> ts
\end{code}

\begin{code}
equs :: String -> [TermPair Var Op Const]
equs = concatMap (equations' . toTerm . snd) . thms
\end{code}

\begin{code}
printEqus :: String -> IO ()
printEqus = mapM_ (\ (x,y) -> print $ apply '=' [x,y]) . equs
\end{code}



%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
